import logging
from typing import Optional, Awaitable, Any, Union

import tornado.web
from tornado import httputil

from .middleware import MiddlewareType

logger = logging.getLogger(__name__)


class MiddlewareHandlerMixin(object):
    def __init__(self, application, request, **kwargs: Any) -> None:
        super().__init__(application, request, **kwargs)

        self.middleware_manager = self.application.middleware_manager

    def prepare(self) -> Optional[Awaitable[None]]:
        """
        在每个请求的最开始的时候被调用, 在 get/post/等方法之前.
        """
        super().prepare()
        return self._process_middlewares(mw_type=MiddlewareType.BEFORE_REQUEST)

    def on_finish(self) -> None:
        """在一个请求结束后被调用."""
        super().on_finish()
        self._process_middlewares(mw_type=MiddlewareType.HANDLE_FINISHED)

    def finish(self, chunk: Union[str, bytes, dict] = None) -> "Future[None]":
        """完成响应后调用."""
        # finish之前可能执行过多次write，反而chunk可能为None
        # 真正的chunk数据在self._write_buffer中，包含历次write的数据
        # 这里将chunk数据write进_write_buffer中，然后将chunk置空
        if chunk:
            self.write(chunk)
            chunk = None
        self._process_middlewares(MiddlewareType.AFTER_RESPONSE, chunk)
        return super().finish(chunk)

    def _process_middlewares(self, mw_type, *args, **kwargs):
        """根据中间件类型执行中间件"""
        self.middleware_manager.run_middleware_type(mw_type, self, *args, **kwargs)


class _PatchHandler(tornado.web.RequestHandler):

    def __init__(self, application, request, **kwargs) -> None:
        super().__init__(application, request, **kwargs)

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass


class BaseHandler(_PatchHandler):
    """A class to collect common handler methods - all other handlers should subclass this one.
    """

    def __init__(self, application: "Application", request: httputil.HTTPServerRequest, **kwargs: Any) -> None:
        super().__init__(application, request, **kwargs)


class RestfulHandler(BaseHandler):
    def __init__(self, application, request, **kwargs: Any) -> None:
        super().__init__(application, request, **kwargs)



import tornado.web

from .exception import ArgumentError
from .middleware import MiddlewareManager, MiddlewareType
from .config import load_config_dir


class Application(tornado.web.Application):
    """Loquat Application"""

    def __init__(self, handlers=None, transforms=None):

        config = load_config_dir()
        default_host = config['default_host']
        app_settings = config['app_settings']

        self._init_middleware(config)

        super(Application, self).__init__(handlers=handlers, default_host=default_host, transforms=transforms,
                                          **app_settings)

    def _init_middleware(self, config):

        default_middlewares = ['middleware.MixinHandlerMiddleware']

        if not isinstance(config['middleware_classes'], list):
            raise ArgumentError('middleware_classes must be list')

        self.middleware_manager = MiddlewareManager()
        self.middleware_manager.register_all(config['middleware_classes'] + default_middlewares)
        self.middleware_manager.run_middleware_type(MiddlewareType.INIT_APPLICATION, self)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)



import logging
import signal
import time

import tornado.escape
import tornado.ioloop
import tornado.locks
import tornado.web
from tornado.options import options

from .log import initialize_logging, NullHandler
from .config import load_config_dir
from .web import Application

# 这里是为了Tornado不初始化日志
logging.getLogger().addHandler(NullHandler())
logger = logging.getLogger(__name__)


def define_server_options():
    """options.define定制"""

    options.define('profile', default='', help='', type=str)
    options.define('config_dir', default='', help='', type=str)


define_server_options()
options.parse_command_line()


class Server(object):
    def __init__(self, application: Application = None, ioloop: object = None, **kwargs):
        """
        @param application: loquat.web.Application及其子类
        @param ioloop: An I/O event loop.
        @param kwargs: 可变参数，可选参数有：
            - config: 配置
            - config_dir： 配置的目录
        """
        self.application = application
        self.ioloop = ioloop
        self.http_server = None

        self._set_config(**kwargs)
        initialize_logging(options=self.config['log'])

    def _set_config(self, **kwargs):
        """设置服务配置

        配置优先级：
            1- 系统运行时传入的配置文件
            2- Server初始化时候的配置
            3- 默认的配置文件
        """

        if kwargs:
            if 'config' in kwargs.keys():
                self.config = kwargs['config']

        # 初始化配置
        config_dir = ''
        if 'config_dir' in kwargs.keys():
            config_dir = kwargs['config_dir']
        if options.config_dir:  # 如果存在 options.config_dir
            config_dir = options.config_dir
        self.config = load_config_dir(config_dir, options.profile)

    def start(self):
        """开始服务"""

        if not self.application:
            app_settings = self.config['app_settings']
            self.application = Application(**app_settings)

        if not self.http_server:
            self.http_server = tornado.httpserver.HTTPServer(self.application, xheaders=True)
        self.http_server.listen(self.config['port'])

        # 处理系统信号
        signal.signal(signal.SIGTERM, self.shutdown_signal_handler)
        signal.signal(signal.SIGINT, self.shutdown_signal_handler)

        logger.info("Application started on port %s." % (self.config['port'],))

        if not self.ioloop:
            self.ioloop = tornado.ioloop.IOLoop.instance()
        self.ioloop.start()

    def shutdown_signal_handler(self, sig, frame):
        """handle shutdown signal"""

        logger.warning('caught signal: %s', sig)

        max_wait_seconds_before_shutdown = 1

        def shutdown():
            logger.info('stopping http server')
            self.http_server.stop()

            logger.info('will shutdown in %s seconds...', max_wait_seconds_before_shutdown)

            deadline = time.time() + max_wait_seconds_before_shutdown

            def stop_loop():
                now = time.time()
                if now < deadline:
                    self.ioloop.add_timeout(now + 1, stop_loop)
                else:
                    logger.info('shutdown')
                    self.ioloop.stop()

            stop_loop()

        self.ioloop.add_callback(shutdown)


def server_start(application: Application):
    server = Server(application)
    server.start()



# coding=utf-8

"""
基于Tornado Log的logger实现.
"""
import logging
import logging.handlers
import os
import sys
from typing import Dict, Any, cast, Union, Optional

bytes_type = bytes
unicode_type = str
basestring_type = str

try:
    import colorama  # type: ignore
except ImportError:
    colorama = None

try:
    import curses
except ImportError:
    curses = None  # type: ignore

_TO_UNICODE_TYPES = (unicode_type, type(None))


def _unicode(value: Union[None, str, bytes]) -> Optional[str]:  # noqa: F811
    """Converts a string argument to a unicode string.

    If the argument is already a unicode string or None, it is returned
    unchanged.  Otherwise it must be a byte string and is decoded as utf8.
    """
    if isinstance(value, _TO_UNICODE_TYPES):
        return value
    if not isinstance(value, bytes):
        raise TypeError("Expected bytes, unicode, or None; got %r" % type(value))
    return value.decode("utf-8")


def _stdout_supports_color() -> bool:
    try:
        if hasattr(sys.stderr, "isatty") and sys.stderr.isatty():
            if curses:
                curses.setupterm()
                if curses.tigetnum("colors") > 0:
                    return True
            elif colorama:
                if sys.stderr is getattr(
                        colorama.initialise, "wrapped_stderr", object()
                ):
                    return True
    except Exception:
        # Very broad exception handling because it's always better to
        # fall back to non-colored logs than to break at startup.
        pass
    return False


def _safe_unicode(s: Any) -> str:
    try:
        return _unicode(s)
    except UnicodeDecodeError:
        return repr(s)


class LogFormatter(logging.Formatter):
    """
    Log formatter used in Tornado.
    """

    DEFAULT_FORMAT = "%(color)s%(asctime)s %(levelname)s --- [%(process)d %(threadName)s] %(name)s - " \
                     "%(filename)s %(lineno)d : %(end_color)s %(message)s"
    DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
    DEFAULT_COLORS = {
        logging.DEBUG: 4,  # Blue
        logging.INFO: 2,  # Green
        logging.WARNING: 3,  # Yellow
        logging.ERROR: 1,  # Red
    }

    def __init__(
            self,
            fmt: str = DEFAULT_FORMAT,
            datefmt: str = DEFAULT_DATE_FORMAT,
            color: bool = True,
            colors: Dict[int, int] = DEFAULT_COLORS,
    ) -> None:
        r"""
        :arg bool color: Enables color support.
        :arg str fmt: Log message format.
          It will be applied to the attributes dict of log records. The
          text between ``%(color)s`` and ``%(end_color)s`` will be colored
          depending on the level if color support is on.
        :arg dict colors: color mappings from logging level to terminal color
          code
        :arg str datefmt: Datetime format.
          Used for formatting ``(asctime)`` placeholder in ``prefix_fmt``.
        """
        logging.Formatter.__init__(self, datefmt=datefmt)
        self._fmt = fmt

        self._colors = {}  # type: Dict[int, str]
        if color and _stdout_supports_color():
            if curses is not None:
                fg_color = curses.tigetstr("setaf") or curses.tigetstr("setf") or b""

                for levelno, code in colors.items():
                    # Convert the terminal control characters from
                    # bytes to unicode strings for easier use with the
                    # logging module.
                    self._colors[levelno] = unicode_type(
                        curses.tparm(fg_color, code), "ascii"
                    )
                self._normal = unicode_type(curses.tigetstr("sgr0"), "ascii")
            else:
                # If curses is not present (currently we'll only get here for
                # colorama on windows), assume hard-coded ANSI color codes.
                for levelno, code in colors.items():
                    self._colors[levelno] = "\033[2;3%dm" % code
                self._normal = "\033[0m"
        else:
            self._normal = ""

    def format(self, record: Any) -> str:
        try:
            message = record.getMessage()
            assert isinstance(message, basestring_type)  # guaranteed by logging
            record.message = _safe_unicode(message)
        except Exception as e:
            record.message = "Bad message (%r): %r" % (e, record.__dict__)

        record.asctime = self.formatTime(record, cast(str, self.datefmt))

        if record.levelno in self._colors:
            record.color = self._colors[record.levelno]
            record.end_color = self._normal
        else:
            record.color = record.end_color = ""

        formatted = self._fmt % record.__dict__

        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            lines = [formatted.rstrip()]
            lines.extend(_safe_unicode(ln) for ln in record.exc_text.split("\n"))
            formatted = "\n".join(lines)
        return formatted.replace("\n", "\n    ")


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


class ExactLogLevelFilter(logging.Filter):
    def __init__(self, level):
        self.__level = level

    def filter(self, log_record):
        return log_record.levelno == self.__level


def _pretty_logging(options: Dict, logger: logging.Logger) -> None:
    # 如果没有设置日志级别
    if options['logging_level'] is None or options['logging_level'].lower() == "none":
        return

    if options['log_file_path']:
        rotate_mode = options['log_rotate_mode']
        if rotate_mode == "size":
            channel = logging.handlers.RotatingFileHandler(
                filename=options['log_file_path'],
                maxBytes=options['log_file_max_size'],
                backupCount=options['log_file_num_backups'],
                encoding="utf-8",
            )  # type: logging.Handler
        elif rotate_mode == "time":
            channel = logging.handlers.TimedRotatingFileHandler(
                filename=options['log_file_path'],
                when=options['log_rotate_when'],
                interval=options['log_rotate_interval'],
                backupCount=options['log_file_num_backups'],
                encoding="utf-8",
            )
        else:
            error_message = (
                    'The value of log_rotate_mode option should be "size" or "time", not "%s".' % rotate_mode
            )
            raise ValueError(error_message)
        channel.setFormatter(LogFormatter(color=False))
        # 添加通过级别过滤
        if options['log_to_level_files']:
            channel.addFilter(ExactLogLevelFilter(logging.getLevelName(options['logging_level'])))
        logger.addHandler(channel)

    if options['log_to_stderr'] or (options['log_to_stderr'] is False and not logger.handlers):
        # Set up color if we are in a tty and curses is installed
        channel = logging.StreamHandler()
        channel.setFormatter(LogFormatter())
        logger.addHandler(channel)


def _log_level_files(default_options, logger):
    """对不同级别的日志分文件记录"""

    log_level_path = {
        'DEBUG': os.path.join(default_options['log_file_path'], 'debug/debug.log'),
        'INFO': os.path.join(default_options['log_file_path'], 'info/info.log'),
        'WARN': os.path.join(default_options['log_file_path'], 'warn/warn.log'),
        'ERROR': os.path.join(default_options['log_file_path'], 'error/error.log')
    }
    log_levels = log_level_path.keys()
    for level in log_levels:
        log_path = os.path.abspath(log_level_path[level])
        if not os.path.exists(os.path.dirname(log_path)):
            os.makedirs(os.path.dirname(log_path))

        default_options.update({'log_file_path': log_path, 'logging_level': level})

        _pretty_logging(options=default_options, logger=logger)


def initialize_logging(logger: logging.Logger = None, options: Dict = None):
    """
    :param logger:
    :param options:
        {
        'log_file_path':'日志文件路径。如果log_level_files==True则为目录，否则为文件路径',
        'logging_level':'日志级别（DEBUG/INFO/WARN/ERROR）',
        'log_to_level_files':'是否按照日志的级别分别记录。默认为False',
        'log_to_stderr':'是否将日志输出发送到stream（如果可能的话，将其着色）。如果未配置其他日志记录，则默认使用stream。',
        'log_file_max_size':'每个文件最大的大小，默认：100 * 1000 * 1000',
        'log_file_num_backups':'要保留的日志文件数',
        'log_rotate_when':'时间间隔的类型（'S', 'M', 'H', 'D', 'W0'-'W6'）',
        'log_rotate_interval':'TimedRotatingFileHandler的interval值',
        'log_rotate_mode':'类型（size/time）'
        }
    :return:
    """

    if logger is None:
        logger = logging.getLogger()

    if options is None:
        options = {}

    default_options = {
        'log_file_path': '',
        'logging_level': 'DEBUG',
        'log_to_level_files': False,
        'log_to_stderr': True,
        'log_file_max_size': 100 * 1000 * 1000,
        'log_file_num_backups': 30,
        'log_rotate_when': 'D',
        'log_rotate_interval': 1,
        'log_rotate_mode': 'time'
    }
    default_options.update(options)

    # 设置日志级别
    logger.setLevel(getattr(logging, default_options['logging_level'].upper()))

    if default_options['log_to_level_files']:
        _log_level_files(default_options, logger)
    else:
        _pretty_logging(options=default_options, logger=logger)



import json
import os

from .util import app_dir


class DeploymentType:
    SOLO = "SOLO"  # 也就是本地模式
    DEV = "DEV"  # 开发环境
    QA = "QA"  # 测试环境
    UAT = "UAT"  # 预生产环境
    PROD = "PROD"  # 生产环境
    dict = {
        SOLO: "SOLO",
        DEV: "DEV",
        QA: "QA",
        UAT: "UAT",
        PROD: "PROD"
    }


default_config = {
    'port': 8000,
    'app_name': '',
    'default_host': '',
    'env': DeploymentType.SOLO,
    'app_settings': {
        'debug': True,
        'xsrf_cookies': True,
    },
    'log': {
        'log_file_path': app_dir + '/app.log',
        'logging_level': 'DEBUG',
        'log_to_level_files': False,
        'log_to_stderr': True,
        'log_file_max_size': 100 * 1000 * 1000,
        'log_file_num_backups': 30,
        'log_rotate_when': 'D',
        'log_rotate_interval': 1,
        'log_rotate_mode': 'time'
    },
    'middleware_classes': []
}


def load_config_dir(config_file_dir='', profile=''):
    _config = default_config
    if not config_file_dir:
        config_file_dir = app_dir + '/conf'  # 默认的配置目录

    config_files = ['application.json']
    if profile:
        config_files.append('application-' + profile.lower() + '.json')

    try:
        files = os.listdir(config_file_dir)
    except FileNotFoundError:
        files = []

    for file in config_files:
        if file in files:
            c = load_config_file(config_file_dir + '/' + file)
            _config = {**_config, **c}

    return _config


def load_config_file(config_file):
    with open(config_file) as f:
        _config = json.load(f)
    return _config



import inspect
import logging
from abc import ABCMeta, abstractmethod

from .exception import ArgumentError
from .util import import_object

logger = logging.getLogger(__name__)


class MiddlewareType:
    INIT_APPLICATION = 'INIT_APPLICATION'  # 在application的init时候调用
    BEFORE_REQUEST = 'BEFORE_REQUEST'  # 在request之前
    AFTER_RESPONSE = 'AFTER_RESPONSE'  # 在response之后
    HANDLE_FINISHED = 'HANDLE_FINISHED'  # 在请求结束之后
    dict = {
        INIT_APPLICATION: "INIT_APPLICATION",
        BEFORE_REQUEST: "BEFORE_REQUEST",
        AFTER_RESPONSE: "AFTER_RESPONSE",
        HANDLE_FINISHED: "HANDLE_FINISHED"
    }


class BaseMiddleware(metaclass=ABCMeta):

    def __init__(self, mw_order=0, mw_type=MiddlewareType.BEFORE_REQUEST):

        self._mw_order = mw_order
        self._mw_type = mw_type

    @property
    def mw_order(self):
        """middleware的执行顺序，值越小越靠前执行
        """
        return self._mw_order

    @mw_order.setter
    def mw_order(self, value):
        if not isinstance(value, int):
            raise ValueError('value must be an integer!')
        if value < 0:
            raise ValueError('value must bigger than 0!')
        self._mw_order = value

    @property
    def mw_type(self):
        """middleware的类型。有：MiddlewareType.FINISHED, MiddlewareType.AFTER_RESPONSE, MiddlewareType.BEFORE_REQUEST
        """
        return self._mw_type

    @mw_type.setter
    def mw_type(self, value):

        if value not in [MiddlewareType.HANDLE_FINISHED, MiddlewareType.AFTER_RESPONSE, MiddlewareType.BEFORE_REQUEST]:
            raise ValueError(
                'value must in [MiddlewareType.FINISHED, MiddlewareType.AFTER_RESPONSE, MiddlewareType.BEFORE_REQUEST]')

        self._mw_type = value

    @abstractmethod
    def should_run(self, component, *args, **kwargs) -> bool:
        """
        是否需要运行中间件。True为运行，False为不运行
        @param component: tornado.web.RequestHandler的子类 或者其他
        @param args:
        @param kwargs:
        """
        pass

    @abstractmethod
    def run(self, component, *args, **kwargs) -> None:
        """
        运行中间件
        @param component: tornado.web.RequestHandler的子类 或者其他
        @param args:
        @param kwargs:
        """
        pass


# class SingletonMiddlewareManagerType(type):
#     _instance_lock = threading.Lock()
#
#     def __call__(cls, *args, **kwargs):
#         if not hasattr(cls, "_middleware_manager"):
#             with SingletonMiddlewareManagerType._instance_lock:
#                 if not hasattr(cls, "_middleware_manager"):
#                     cls._middleware_manager = super(SingletonMiddlewareManagerType, cls).__call__(*args, **kwargs)
#         return cls._middleware_manager


class MiddlewareManager(object):

    def __init__(self) -> None:
        self.init_application_list = []
        self.before_request_list = []
        self.after_response_list = []
        self.handle_finished_list = []

    def register(self, middleware):
        """注册中间件"""
        if not middleware:
            return
        if isinstance(middleware, str):
            middleware = import_object(middleware)
        if not issubclass(middleware, BaseMiddleware):
            cl = [c for c in inspect.getmro(middleware) if c.__name__ == 'BaseMiddleware']
            if len(cl) < 1:
                raise ArgumentError("middleware must be a subclass of BaseMiddleware")

        instance = middleware()
        mw_type = instance.mw_type
        if mw_type is MiddlewareType.INIT_APPLICATION:
            self.init_application_list.append(instance)
            self.init_application_list = sorted(self.init_application_list, key=lambda m: m.mw_order)
        elif mw_type is MiddlewareType.BEFORE_REQUEST:
            self.before_request_list.append(instance)
            self.before_request_list = sorted(self.before_request_list, key=lambda m: m.mw_order)
        elif mw_type is MiddlewareType.AFTER_RESPONSE:
            self.after_response_list.append(instance)
            self.after_response_list = sorted(self.after_response_list, key=lambda m: m.mw_order)
        elif mw_type is MiddlewareType.HANDLE_FINISHED:
            self.handle_finished_list.append(instance)
            self.handle_finished_list = sorted(self.handle_finished_list, key=lambda m: m.mw_order)

    def register_all(self, middlewares):
        """注册多个中间件"""
        if not middlewares:
            return

        for middleware in middlewares:
            self.register(middleware)

    def run_middleware_type(self, mw_type: str, component: object, *args, **kwargs) -> tuple:
        """
        根据中间件的类型运行中间件列表
        @param mw_type:
        @param component:
        @param args:
        @param kwargs:
        @return:
        """
        mws = []
        if mw_type is MiddlewareType.INIT_APPLICATION:
            mws = self.init_application_list
        elif mw_type is MiddlewareType.BEFORE_REQUEST:
            mws = self.before_request_list
        elif mw_type is MiddlewareType.AFTER_RESPONSE:
            mws = self.after_response_list
        elif mw_type is MiddlewareType.HANDLE_FINISHED:
            mws = self.handle_finished_list

        res_list = []
        for m in mws:
            res = self.run_middleware(m, component, *args, **kwargs)
            res_list.append(res)
        return tuple(res_list)

    def run_middleware(self, middleware, component, *args, **kwargs):
        """
        运行中间件
        @param middleware:
        @param component:
        @param args:
        @param kwargs:
        @return:
        """
        try:
            should_run_res = middleware.should_run(component, *args, **kwargs)
            if should_run_res:
                res = middleware.run(component, *args, **kwargs)
                return res
        except Exception as e:
            logger.exception(e)
            raise e


# 以下为系统默认的中间件
class MixinHandlerMiddleware(BaseMiddleware):
    """
    组合Handler的中间件。根据application进行判断是否要组合
    """

    def __init__(self, mw_order=0, mw_type=MiddlewareType.INIT_APPLICATION):
        super().__init__(mw_order, mw_type)

        self.mm_should = False  # 是否组合MiddlewareHandlerMixin

    def should_run(self, application, *args, **kwargs) -> bool:
        m = application.middleware_manager
        # 如果存在中间件，那么就设置为true
        mm_should = len(m.before_request_list) > 0 or len(m.after_response_list) > 0 or len(m.handle_finished_list) > 0
        self.mm_should = mm_should
        return self.mm_should

    def run(self, application, *args, **kwargs) -> None:
        if self.mm_should:
            from handler import BaseHandler, MiddlewareHandlerMixin
            # 一定要将BaseHandler.__bases__ 写后面，这样才能覆盖，使用MiddlewareHandlerMixin中的方法
            BaseHandler.__bases__ = (MiddlewareHandlerMixin,) + BaseHandler.__bases__
